# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`stackdriver.py`
Stackdriver Monitoring exporter class.
"""
import logging

from prometheus_client import CollectorRegistry, Gauge, push_to_gateway
from prometheus_client.exposition import basic_auth_handler, default_handler

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)
DEFAULT_PUSHGATEWAY_JOB = "slo-generator"

class PrometheusExporter(MetricsExporter):
    """Prometheus exporter class."""
    REQUIRED_FIELDS = ['url']
    OPTIONAL_FIELDS = ['job', 'username', 'password']

    def __init__(self):
        self.username = None
        self.password = None

    def export_metric(self, data):
        """Export data to Prometheus.

        Args:
            data (dict): Metric data.

        Returns:
            object: Prometheus API result.
        """
        self.create_timeseries(data)

    def create_timeseries(self, data):
        """Create Prometheus timeseries.

        Args:
            data (dict): Metric data.

        Returns:
            object: Metric descriptor.
        """
        name = data['name']
        description = data['description']
        prometheus_push_url = data['url']
        prometheus_push_job_name = data.get('job', DEFAULT_PUSHGATEWAY_JOB)
        value = data['value']

        # Write timeseries w/ metric labels.
        labels = data['labels']
        registry = CollectorRegistry()
        gauge = Gauge(name,
                      description,
                      registry=registry,
                      labelnames=labels.keys())
        gauge.labels(*labels.values()).set(value)

        # Handle headers
        handler = default_handler
        if 'username' in data and 'password' in data:
            self.username = data['username']
            self.password = data['password']
            handler = PrometheusExporter.auth_handler

        return push_to_gateway(prometheus_push_url,
                               job=prometheus_push_job_name,
                               grouping_key=labels,
                               registry=registry,
                               handler=handler)

    def auth_handler(self, url, method, timeout, headers, data):
        """Handles authentication for pushing to Prometheus gateway.

        Args:
            url (str): Prometheus gateway URL.
            method (str): Prometheus query method.
            timeout (int): Prometheus timeout.
            headers (dict): Headers.
            data (dict): Data to send.

        Returns:
            func: Auth handler function.
        """
        username = self.username
        password = self.password
        return basic_auth_handler(url, method, timeout, headers, data, username,
                                  password)



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from mock import patch

from slo_generator.cli import cli, parse_args

from .test_stubs import CTX, mock_sd

cwd = os.path.dirname(os.path.abspath(__file__))
root = os.path.dirname(os.path.dirname(cwd))


class TestCLI(unittest.TestCase):
    def setUp(self):
        for k, v in CTX.items():
            os.environ[k] = v
        slo_config = f'{root}/samples/stackdriver/slo_gae_app_availability.yaml'
        eb_policy = f'{root}/samples/error_budget_policy.yaml'
        self.slo_config = slo_config
        self.eb_policy = eb_policy

    def test_parse_args(self):
        args = parse_args([
            '--slo-config', self.slo_config, '--error-budget-policy',
            self.eb_policy, '--export'
        ])
        self.assertEqual(args.slo_config, self.slo_config)
        self.assertEqual(args.error_budget_policy, self.eb_policy)
        self.assertEqual(args.export, True)

    @patch('google.api_core.grpc_helpers.create_channel',
           return_value=mock_sd(8))
    def test_cli(self, mock):
        args = parse_args(['-f', self.slo_config, '-b', self.eb_policy])
        all_reports = cli(args)
        len_first_report = len(all_reports[self.slo_config])
        self.assertIn(self.slo_config, all_reports.keys())
        self.assertEqual(len_first_report, 4)

    @patch('google.api_core.grpc_helpers.create_channel',
           return_value=mock_sd(40))
    def test_cli_folder(self, mock):
        args = parse_args(
            ['-f', f'{root}/samples/stackdriver', '-b', self.eb_policy])
        all_reports = cli(args)
        len_first_report = len(all_reports[self.slo_config])
        self.assertIn(self.slo_config, all_reports.keys())
        self.assertEqual(len_first_report, 4)

    def test_cli_no_config(self):
        args = parse_args([
            '-f', f'{root}/samples', '-b',
            f'{root}/samples/error_budget_policy.yaml'
        ])
        all_reports = cli(args)
        self.assertEqual(all_reports, {})


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`base.py`
Base exporter abstract classes.
"""
import logging
import warnings
from abc import ABCMeta, abstractmethod

LOGGER = logging.getLogger(__name__)

DEFAULT_METRIC_LABELS = [
    'error_budget_policy_step_name', 'service_name', 'feature_name', 'slo_name',
    'metadata'
]

DEFAULT_METRICS = [
    {
        'name': 'error_budget_burn_rate',
        'description': 'Speed at which the error budget is consumed.',
        'labels': DEFAULT_METRIC_LABELS
    },
    {
        'name': 'alerting_burn_rate_threshold',
        'description': 'Error Budget burn rate threshold.',
        'labels': DEFAULT_METRIC_LABELS
    },
    {
        'name': 'events_count',
        'description': 'Number of events',
        'labels': DEFAULT_METRIC_LABELS + [
            'good_events_count', 'bad_events_count']
    },
    {
        'name': 'sli_measurement',
        'description': 'Service Level Indicator.',
        'labels': DEFAULT_METRIC_LABELS
    },
    {
        'name': 'slo_target',
        'description': 'Service Level Objective target.',
        'labels': DEFAULT_METRIC_LABELS
    },
]

class MetricsExporter:
    """Abstract class to export metrics to different backends. Common format
    for YAML configuration to configure which metrics should be exported."""
    __metaclass__ = ABCMeta

    def export(self, data, **config):
        """Export metric data. Loops through metrics config and calls the child
        class `export_metric` method.

        Args:
            data (dict): SLO Report data.
            config (dict): Exporter config.

        Returns:
            list: List of exporter responses.
        """
        metrics = config.get('metrics', DEFAULT_METRICS)
        required_fields = getattr(self, 'REQUIRED_FIELDS', [])
        optional_fields = getattr(self, 'OPTIONAL_FIELDS', [])
        all_data = []
        LOGGER.debug(
            f'Exporting {len(metrics)} metrics with {self.__class__.__name__}')
        for metric_cfg in metrics:
            if isinstance(metric_cfg, str): # short form
                metric_cfg = {
                    'name': metric_cfg,
                    'alias': metric_cfg,
                    'description': "",
                    'labels': DEFAULT_METRIC_LABELS
                }
            if metric_cfg['name'] == 'error_budget_burn_rate':
                metric_cfg = MetricsExporter.use_deprecated_fields(
                    config=config,
                    metric=metric_cfg)
            metric = metric_cfg.copy()
            fields = {
                key: value for key, value in config.items()
                if key in required_fields or key in optional_fields
            }
            metric.update(fields)
            metric = self.build_metric(data, metric)
            name = metric['name']
            labels = metric['labels']
            labels_str = ', '.join([f'{k}={v}' for k, v in labels.items()])
            LOGGER.info(
                f'Exporting "{name}" with {len(labels.keys())} labels: '
                f'{labels_str}...')
            ret = self.export_metric(metric)
            metric_info = {
                k: v for k, v in metric.items()
                if k in ['name', 'alias', 'description', 'labels']
            }
            response = {
                'response': ret,
                'metric': metric_info
            }
            if ret and 'error' in ret:
                LOGGER.error(response)
            all_data.append(response)
        return all_data

    def build_metric(self, data, metric):
        """Build a metric from current data and metric configuration.
        Set the metric value labels and eventual alias.

        Args:
            data (dict): SLO Report data.
            metric (dict): Metric configuration.

        Returns:
            dict: Metric configuration.
        """
        name = metric['name']
        prefix = getattr(self, 'METRIC_PREFIX', None)

        # Set value + timestamp
        metric['value'] = data[name]
        metric['timestamp'] = data['timestamp']

        # Set metric data labels
        labels = metric.get('labels', DEFAULT_METRIC_LABELS).copy()
        additional_labels = metric.get('additional_labels', [])
        labels.extend(additional_labels)
        labels = MetricsExporter.build_data_labels(data, labels)
        metric['labels'] = labels

        # Use metric alias (mapping)
        if 'alias' in metric:
            metric['name'] = metric['alias']

        if prefix and not metric['name'].startswith(prefix):
            metric['name'] = prefix + metric['name']

        # Set description
        metric['description'] = metric.get('description', "")
        return metric

    @staticmethod
    def build_data_labels(data, labels):
        """Build data labels. Also handle nested labels (depth=1).

        Args:
            data (dict): SLO Report data.
            labels (list): Label keys.

        Returns:
            dict: Data labels.
        """
        data_labels = {}
        nested_labels = [
            label for label in labels
            if label in data and isinstance(data[label], dict)
        ]
        flat_labels = [
            label for label in labels
            if label in data and not isinstance(data[label], dict)
        ]
        for label in nested_labels:
            data_labels.update({k: str(v) for k, v in data[label].items()})
        for label in flat_labels:
            data_labels.update({label: str(data[label])})
        LOGGER.debug(f'Data labels: {data_labels}')
        return data_labels

    @staticmethod
    def use_deprecated_fields(config, metric):
        """Old format to new format with DeprecationWarning for 2.0.0.

        Update error_budget_burn_rate metric with `metric_type`,
        `metric_labels`, and `metric_description`.

        Args:
            config (dict): Exporter config.
            metric (dict): Metric config.

        Returns:
            list: List of metrics to export.
        """
        old_metric_type = config.get('metric_type')
        old_metric_labels = config.get('metric_labels')
        old_metric_description = config.get('metric_description')
        if old_metric_type:
            metric['alias'] = old_metric_type
            warnings.warn(
                '`metric_type` will be deprecated in favor of `metrics` '
                'in version 2.0.0, ', FutureWarning)
        if old_metric_labels:
            metric['labels'] = old_metric_labels
            warnings.warn(
                '`metric_labels` will be deprecated in favor of `metrics` '
                'in version 2.0.0, ', FutureWarning)
        if old_metric_description:
            warnings.warn(
                '`metric_description` will be deprecated in favor of `metrics` '
                'in version 2.0.0, ', FutureWarning)
            metric['description'] = old_metric_description
        return metric

    @abstractmethod
    def export_metric(self, data):
        """Abstract method to export a metric. Should be implemented by children
        classes."""
        raise NotImplementedError



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""stubs.py

Stubs for mocking backends and exporters.
"""
import copy
import json
import os
import sys
import time
from types import ModuleType

from google.cloud.monitoring_v3.proto import metric_service_pb2
from slo_generator.utils import list_slo_configs, parse_config

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
SAMPLE_DIR = os.path.join(os.path.dirname(os.path.dirname(TEST_DIR)),
                          "samples/")

CTX = {
    'PUBSUB_PROJECT_ID': 'fake',
    'PUBSUB_TOPIC_NAME': 'fake',
    'GAE_PROJECT_ID': 'fake',
    'GAE_MODULE_ID': 'fake',
    'GKE_MESH_UID': 'fake',
    'GKE_PROJECT_ID': 'fake',
    'GKE_CLUSTER_NAME': 'fake',
    'GKE_LOCATION': 'fake',
    'GKE_SERVICE_NAMESPACE': 'fake',
    'GKE_SERVICE_NAME': 'fake',
    'LB_PROJECT_ID': 'fake',
    'PROMETHEUS_URL': 'http://localhost:9090',
    'PROMETHEUS_PUSHGATEWAY_URL': 'http://localhost:9091',
    'ELASTICSEARCH_URL': 'http://localhost:9200',
    'STACKDRIVER_HOST_PROJECT_ID': 'fake',
    'STACKDRIVER_LOG_METRIC_NAME': 'fake',
    'BIGQUERY_PROJECT_ID': 'fake',
    'BIGQUERY_TABLE_ID': 'fake',
    'BIGQUERY_DATASET_ID': 'fake',
    'BIGQUERY_TABLE_NAME': 'fake',
    'DATADOG_API_KEY': 'fake',
    'DATADOG_APP_KEY': 'fake',
    'DATADOG_SLO_ID': 'fake',
    'DYNATRACE_API_URL': 'fake',
    'DYNATRACE_API_TOKEN': 'fake'
}

CUSTOM_BACKEND_CODE = """
class DummyBackend:
    def __init__(self, client=None, **config):
        self.good_events = config.get('good_events', None)
        self.bad_events = config.get('bad_events', None)
        self.sli_value = config.get('sli', None)

    def good_bad_ratio(self, timestamp, window, slo_config):
        return (self.good_events, self.bad_events)

    def sli(self, timestamp, window, slo_config):
        return self.sli_value
"""

FAIL_EXPORTER_CODE = """
from slo_generator.exporters.base import MetricsExporter
class FailExporter(MetricsExporter):
    def export_metric(self, data):
        raise ValueError("Oops !")
"""

def add_dynamic(name, code, type):
    """Dynamically add a backend or exporter to slo-generator.

    Args:
        name (str): Name of backend / exporter.
        code (str): Backend / exporter code.
        type (str): 'backends' or 'exporters'.
    """
    mod = ModuleType(name)
    module_name = f'slo_generator.{type}.{name}'
    sys.modules[module_name] = mod
    exec(code, mod.__dict__)


# Add backends / exporters for testing purposes
add_dynamic('dummy', CUSTOM_BACKEND_CODE, 'backends')
add_dynamic('fail', FAIL_EXPORTER_CODE, 'exporters')


CUSTOM_BASE_CONFIG = {
    "service_name": "test",
    "feature_name": "test",
    "slo_name": "test",
    "slo_description": "Test dummy backend",
    "slo_target": 0.99,
    "backend": {
        "class": "Dummy",
    }
}

CUSTOM_STEP = {
    "error_budget_policy_step_name": "1 hour",
    "measurement_window_seconds": 3600,
    "alerting_burn_rate_threshold": 1,
    "urgent_notification": True,
    "overburned_consequence_message": "Page to defend the SLO",
    "achieved_consequence_message": "Last hour on track"
}

CUSTOM_TESTS = {
    "enough_events": {
        'method': 'good_bad_ratio',
        'good_events': 5,
        'bad_events': 5,
    },
    "no_good_events": {
        'method': 'good_bad_ratio',
        'good_events': -1,
        'bad_events': 15,
    },
    "no_bad_events": {
        'method': 'good_bad_ratio',
        'good_events': 15,
        'bad_events': -1,
    },
    "valid_sli_value": {
        'method': 'sli',
        'good_events': -1,
        'bad_events': -1,
        'sli': 0.991
    },
    "no_events": {
        'method': 'good_bad_ratio',
        'good_events': 0,
        'bad_events': 0,
    },
    "no_good_bad_events": {
        'method': 'good_bad_ratio',
        'good_events': -1,
        'bad_events': -1,
    },
    "not_enough_events": {
        'method': 'good_bad_ratio',
        'good_events': 5,
        'bad_events': 4,
    },
    "no_sli_value": {
        'method': 'sli',
        'good_events': -1,
        'bad_events': -1,
    },
    "no_backend_response_sli": {
        'method': 'sli',
        'sli': None
    },
    "no_backend_response_ratio": {
        'method': 'good_bad_ratio',
        'good_events': None,
        'bad_events': None,
    },
    'invalid_backend_response_type': {
        'method': 'good_bad_ratio',
        'good_events': {
            'data': {
                'value': 30
            }
        },
        'bad_events': {
            'data': {
                'value': 400
            }
        },
        'sli': None
    }
}


def mock_slo_report(key):
    """Mock SLO report config with edge cases contained in CUSTOM_TESTS.

    Args:
        key (str): Key identifying which config to pick from CUSTOM_TESTS.

    Returns:
        dict: Dict configuration for SLOReport class.
    """
    config = copy.deepcopy(CUSTOM_BASE_CONFIG)
    config["backend"].update(CUSTOM_TESTS[key])
    timestamp = time.time()
    return {
        "config": config,
        "step": CUSTOM_STEP,
        "timestamp": timestamp,
        "client": None,
        "delete": False
    }


# pylint: disable=too-few-public-methods
class MultiCallableStub:
    """Stub for the grpc.UnaryUnaryMultiCallable interface."""
    def __init__(self, method, channel_stub):
        self.method = method
        self.channel_stub = channel_stub

    # pylint: disable=inconsistent-return-statements
    def __call__(self, request, timeout=None, metadata=None, credentials=None):
        self.channel_stub.requests.append((self.method, request))

        response = None
        if self.channel_stub.responses:
            response = self.channel_stub.responses.pop()

        if isinstance(response, Exception):
            raise response

        if response:
            return response


# pylint: disable=R0903
class ChannelStub:
    """Stub for the grpc.Channel interface."""
    def __init__(self, responses=[]):
        self.responses = responses
        self.requests = []

    # pylint: disable=C0116,W0613
    def unary_unary(self,
                    method,
                    request_serializer=None,
                    response_deserializer=None):
        return MultiCallableStub(method, self)


def mock_grpc_stub(response, proto_method, nresp=1):
    """Fakes gRPC response channel for the proto_method passed.

    Args:
        response (dict): Expected response.
        nresp (int): Number of expected responses.

    Returns:
        ChannelStub: Mocked gRPC channel stub.
    """
    expected_response = proto_method(**response)
    channel = ChannelStub(responses=[expected_response] * nresp)
    return channel


def mock_sd(nresp=1):
    """Fake Stackdriver Monitoring API response for the ListTimeSeries endpoint.

    Args:
        nresp (int): Number of responses to add to response.

    Returns:
        ChannelStub: Mocked gRPC channel stub.
    """
    timeserie = load_fixture('time_series_proto.json')
    response = {"next_page_token": "", "time_series": [timeserie]}
    return mock_grpc_stub(
        response=response,
        proto_method=metric_service_pb2.ListTimeSeriesResponse,
        nresp=nresp)


# pylint: disable=W0613,R1721
def mock_prom(self, metric):
    """Fake Prometheus query response.

    Args:
        metric (dict): Input metric query.

    Returns:
        dict: Fake response.
    """
    data = {
        'data': {
            'result': [{
                'values': [x for x in range(5)],
                'value': [0, 1]
            }]
        }
    }
    return json.dumps(data)


# pylint: disable=W0613
def mock_es(self, index, body):
    """Fake ElasticSearch response.

    Args:
        index (str): Index.
        body (dict): Query body.

    Returns:
        dict: Fake response.
    """
    return {'hits': {'total': {'value': 120}}}


def mock_dd_metric_query(*args, **kwargs):
    """Mock Datadog response for datadog.api.Metric.query."""
    return load_fixture('dd_timeseries.json')


def mock_dd_slo_history(*args, **kwargs):
    """Mock Datadog response for datadog.api.ServiceLevelObjective.history."""
    return load_fixture('dd_slo_history.json')


def mock_dd_slo_get(*args, **kwargs):
    """Mock Datadog response for datadog.api.ServiceLevelObjective.get."""
    return load_fixture('dd_slo.json')


def mock_dd_metric_send(*args, **kwargs):
    """Mock Datadog response for datadog.api.Metric.send."""
    return load_fixture('dd_success.json')


def mock_dt(*args, **kwargs):
    """Mock Dynatrace response."""
    if args[0] == 'get' and args[1] == 'timeseries':
        return load_fixture('dt_metric_get.json')

    elif args[0] == 'get' and args[1] == 'metrics/query':
        return load_fixture('dt_timeseries_get.json')

    elif args[0] == 'post' and args[1] == 'entity/infrastructure/custom':
        return load_fixture('dt_metric_send.json')

    elif args[0] == 'put' and args[1] == 'timeseries':
        return {}

def mock_dt_errors(*args, **kwargs):
    """Mock Dynatrace response with errors."""
    if args[0] == 'get' and args[1] == 'timeseries':
        return load_fixture('dt_metric_get.json')

    elif args[0] == 'get' and args[1] == 'metrics/query':
        return load_fixture('dt_timeseries_get.json')

    elif args[0] == 'post' and args[1] == 'entity/infrastructure/custom':
        return load_fixture('dt_error_rate.json')

    elif args[0] == 'put' and args[1] == 'timeseries':
        return load_fixture('dt_error_rate.json')

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def dotize(data):
    """Transform dict to class instance with attribute access.

    Args:
        data (dict): Input dict.

    Returns:
        dotdict: Dotdict equivalent.
    """
    data = dotdict(data)
    for k, v in data.items():
        if isinstance(v, dict):
            data[k] = dotdict(v)
    return data


class mock_ssm_client:
    """Fake Service Monitoring API client."""
    def __init__(self):
        self.services = [dotize(s) for s in load_fixture('ssm_services.json')]
        self.service_level_objectives = [
            dotize(slo) for slo in load_fixture('ssm_slos.json')
        ]

    def project_path(self, project_id):
        return f'projects/{project_id}'

    def service_path(self, project_id, service_id):
        project_path = self.project_path(project_id)
        return f'{project_path}/services/{service_id}'

    def create_service(self, parent, service, service_id=None):
        return self.services[0]

    def list_services(self, parent):
        return self.services

    def delete_service(self, name):
        return None

    def create_service_level_objective(self,
                                       parent,
                                       service_level_objective,
                                       service_level_objective_id=None):
        return self.service_level_objectives[0]

    def update_service_level_objective(self, service_level_objective):
        return self.service_level_objectives[0]

    def list_service_level_objectives(self, parent):
        return self.service_level_objectives

    def delete_service_level_objective(self, name):
        return None

    @staticmethod
    def to_json(data):
        return data


def load_fixture(filename, **ctx):
    """Load a fixture from the test/fixtures/ directory and replace context
    environmental variables in it.

    Args:
        filename (str): Filename of the fixture to load.
        ctx (dict): Context dictionary (env variables).

    Returns:
        dict: Loaded fixture.
    """
    filename = os.path.join(TEST_DIR, "fixtures/", filename)
    return parse_config(filename, ctx)


def load_sample(filename, **ctx):
    """Load a sample from the samples/ directory and replace context
    environmental variables in it.

    Args:
        filename (str): Filename of the fixture to load.
        ctx (dict): Context dictionary (env variables).

    Returns:
        dict: Loaded sample.
    """
    filename = os.path.join(SAMPLE_DIR, filename)
    return parse_config(filename, ctx)


def load_slo_samples(folder_path, **ctx):
    """List and load all SLO samples from folder path.

    Args:
        folder_path (str): Folder path to load SLO configs from.
        ctx (dict): Context for env variables.

    Returns:
        list: List of loaded SLO configs.
    """
    return [
        load_sample(filename, **ctx)
        for filename in list_slo_configs(f'{SAMPLE_DIR}/{folder_path}')
    ]



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`stackdriver.py`
Stackdriver Monitoring exporter class.
"""
import logging

import google.api_core.exceptions
from google.cloud import monitoring_v3

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)

class StackdriverExporter(MetricsExporter):
    """Stackdriver Monitoring exporter class."""
    METRIC_PREFIX = "custom.googleapis.com/"
    REQUIRED_FIELDS = ['project_id']

    def __init__(self):
        self.client = monitoring_v3.MetricServiceClient()

    def export_metric(self, data):
        """Export metric to Stackdriver Monitoring. Create metric descriptor if
        it doesn't exist.

        Args:
            data (dict): Data to send to Stackdriver Monitoring.
            project_id (str): Stackdriver Monitoring project id.

        Returns:
            object: Stackdriver Monitoring API result.
        """
        if not self.get_metric_descriptor(data):
            self.create_metric_descriptor(data)
        self.create_timeseries(data)

    def create_timeseries(self, data):
        """Create Stackdriver Monitoring timeseries.

        Args:
            data (dict): Metric data.

        Returns:
            object: Metric descriptor.
        """
        labels = data['labels']
        series = monitoring_v3.types.TimeSeries()
        series.metric.type = data['name']
        for key, value in labels.items():
            series.metric.labels[key] = value
        series.resource.type = 'global'

        # Create a new data point.
        point = series.points.add()

        # Define end point timestamp.
        timestamp = data['timestamp']
        point.interval.end_time.seconds = int(timestamp)
        point.interval.end_time.nanos = int(
            (timestamp - point.interval.end_time.seconds) * 10**9)

        # Set the metric value.
        point.value.double_value = data['value']

        # Record the timeseries to Stackdriver Monitoring.
        project = self.client.project_path(data['project_id'])
        result = self.client.create_time_series(project, [series])
        labels = series.metric.labels
        LOGGER.debug(
            f"timestamp: {timestamp} value: {point.value.double_value}"
            f"{labels['service_name']}-{labels['feature_name']}-"
            f"{labels['slo_name']}-{labels['error_budget_policy_step_name']}")
        return result

    def get_metric_descriptor(self, data):
        """Get Stackdriver Monitoring metric descriptor.

        Args:
            data (dict): Metric data.

        Returns:
            object: Metric descriptor (or None if not found).
        """
        descriptor = self.client.metric_descriptor_path(data['project_id'],
                                                        data['name'])
        try:
            return self.client.get_metric_descriptor(descriptor)
        except google.api_core.exceptions.NotFound:
            return None

    def create_metric_descriptor(self, data):
        """Create Stackdriver Monitoring metric descriptor.

        Args:
            data (dict): Metric data.

        Returns:
            object: Metric descriptor.
        """
        project = self.client.project_path(data['project_id'])
        descriptor = monitoring_v3.types.MetricDescriptor()
        descriptor.type = data['name']
        descriptor.metric_kind = (
            monitoring_v3.enums.MetricDescriptor.MetricKind.GAUGE)
        descriptor.value_type = (
            monitoring_v3.enums.MetricDescriptor.ValueType.DOUBLE)
        descriptor.description = data['description']
        self.client.create_metric_descriptor(project, descriptor)
        return descriptor



# Copyright 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from io import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Package metadata.
name = "slo-generator"
description = "SLO Generator"
version = "1.4.0"
# Should be one of:
# 'Development Status :: 3 - Alpha'
# 'Development Status :: 4 - Beta'
# 'Development Status :: 5 - Production/Stable'
release_status = "Development Status :: 3 - Alpha"
dependencies = [
    'google-api-python-client < 2.0.0', 'oauth2client',
    'google-cloud-monitoring < 2.0.0', 'google-cloud-pubsub==1.7.0',
    'google-cloud-bigquery < 3.0.0', 'prometheus-http-client',
    'prometheus-client', 'pyyaml', 'opencensus', 'elasticsearch',
    'python-dateutil', 'datadog', 'retrying==1.3.3'
]
extras = {}

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name=name,
      version=version,
      description=description,
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Google Inc.',
      author_email='ocervello@google.com',
      license='Apache 2.0',
      packages=find_packages(exclude=['contrib', 'docs', 'tests']),
      classifiers=[
          release_status,
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Build Tools',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
      ],
      keywords='slo sli generator gcp',
      install_requires=dependencies,
      entry_points={
          'console_scripts': ['slo-generator=slo_generator.cli:main'],
      },
      python_requires='>=3.4')



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`datadog.py`
Datadog exporter class.
"""
import logging

import datadog

from .base import MetricsExporter

LOGGER = logging.getLogger(__name__)
logging.getLogger('datadog.api').setLevel(logging.ERROR)

DEFAULT_API_HOST = "https://api.datadoghq.com"

# pylint: disable=too-few-public-methods
class DatadogExporter(MetricsExporter):
    """Datadog exporter class.

    Args:
        client (obj, optional): Existing Datadog client to pass.
        api_key (str): Datadog API key.
        app_key (str): Datadog APP key.
        kwargs (dict): Extra arguments to pass to initialize function.
    """
    REQUIRED_FIELDS = ['api_key', 'app_key']
    OPTIONAL_FIELDS = ['api_host']

    def export_metric(self, data):
        """Export a metric to Datadog.

        Args:
            data (dict): Metric data.

        Raises:
            DatadogError (object): Datadog exception object.
        """
        options = {
            'api_key': data['api_key'],
            'app_key': data['app_key'],
            'api_host': data.get('api_host', DEFAULT_API_HOST)
        }
        datadog.initialize(**options)
        client = datadog.api
        timestamp = data['timestamp']
        tags = data['labels']
        name = data['name']
        value = data['value']
        return client.Metric.send(metric=name,
                                  points=[(timestamp, value)],
                                  tags=tags)
